# sourcery skip: raise-specific-error
from django.apps import apps
from django.db import models
from .base_enum import Values
from .exception import ModelError
from .middleware import get_current_user
from .validators import validate_postive
from django.db.models import Q
from django.db.models.constraints import UniqueConstraint



try:
   from django.conf import settings
except ImportError as e:
   raise Exception("settings.py file is required to run this package") from e
from django.conf import settings


## CORE CONFIGURATIONS CLASSES ##


class Flowmodel(models.Model):
    description = models.CharField(max_length=255)

    def __str__(self):
        return self.description

    class Meta:
        ordering = ['id']
        verbose_name_plural = "1. Flowmodel"



class Party(models.Model):
    description = models.CharField(max_length=255 , unique = True)

    def __str__(self):
        return self.description

    class Meta:
        ordering = ['id']
        verbose_name_plural = "2. Party"


class TransitionManager(models.Model):
    t_id = models.IntegerField()
    type = models.CharField(max_length = 255 )
    sub_sign = models.IntegerField(default = 0 , editable = False)
    in_progress = models.BooleanField(default = False)
    created_date = models.DateTimeField(auto_now_add=True)
    
    # default 1 for initial submit and INITIAL_SIGN process
    
    def __str__(self):
        return "{0}_{1}".format(self.type.capitalize() , self.t_id)

    @property
    def sign_reset(self):
        TransitionManager.objects.filter(t_id = self.t_id , type = self.type , id = self.id).update(
            sub_sign = 0 , in_progress = False
        )
        return None
    
    class Meta:
        ordering = ['id']
        verbose_name_plural = "4. TransitionManager"


class SignList(models.Model):
    sign_id = models.IntegerField(validators=[validate_postive],editable = False,blank=True, null=True)
    name = models.CharField(max_length = 255 , unique = True )
    sub_action_name = models.CharField(max_length = 255 , unique = True)

    def __str__(self):
        return self.name
     
    class Meta:
        ordering = ['id']
        verbose_name_plural = "5. Signatures"

    def save(self,*args, **kwargs):
       super(SignList, self).save(*args, **kwargs)
       sign_list_counter = SignList.objects.all().count()
       for i in range(sign_list_counter):
          SignList.objects.filter(id = i+1).update(sign_id = i)
       return None
    



def party_models():
    try: 
        return settings.FINFLO['PARTY_MODEL'][0] or Party
    except Exception:
        return Party
    

class States(models.Model):
    description = models.CharField(max_length=255)
    
    class Meta:
        ordering = ['id']
        verbose_name_plural = "6. States"

    def __str__(self): 
        return self.description
    
    def save(self, *args, **kwargs):
        self.description = self.description.capitalize()
        return super(States, self).save(*args, **kwargs) 


class Action(models.Model):
    party = models.ForeignKey(party_models() , models.CASCADE , blank = True , null = True , help_text = 'this field is optional' , related_name = 'user_party_model')
    description = models.CharField(max_length=255 , blank = True , null=True , help_text = 'e.g., SUBMIT , DELETE') 
    model = models.ForeignKey(Flowmodel , on_delete = models.CASCADE , blank = True , null = True )
    from_state = models.ForeignKey(States , on_delete= models.DO_NOTHING , related_name = 'action_from_state' , blank = True , null = True , help_text = "initial from state for transition ")
    to_state = models.ForeignKey(States , on_delete= models.DO_NOTHING , blank = True , null = True , help_text = "final state for the transition to take place")
    intermediator = models.BooleanField(default = False)
    from_party = models.ForeignKey(party_models() , models.CASCADE , blank = True , null = True , help_text = 'this field is optional' , related_name = 'from_transition_party_type')
    to_party = models.ForeignKey(party_models() , models.CASCADE , blank = True , null = True , help_text = 'this field is optional' , related_name = 'to_transition_party_type')
    stage_required = models.ForeignKey(SignList , on_delete= models.DO_NOTHING ,blank = True , null = True ,help_text = "this field is optional , IMPORTANT : if INITIAL_SIGN means initial_transition")
    sign_required = models.IntegerField(default=0,editable = False,help_text = "IMPORTANT : if 0 means initial_transition " , blank=True, null=True )

 
    def save(self, *args, **kwargs):
        self.description = self.description.upper()
        try:
            sign_len = SignList.objects.get(name = self.stage_required.name )
            self.sign_required = sign_len.sign_id
        except Exception:
            self.sign_required = 0
        return super(Action,self).save(*args, **kwargs)

    
    
    def __str__(self):
        return self.description

    class Meta:
        ordering = ['id']
        verbose_name_plural = "3. Action"
        # constraints = [
        #     UniqueConstraint(fields=['description', 'model', 'party'],
        #                      name='unique_with_optional'),
        #     UniqueConstraint(fields=['description', 'model'],
        #                      condition=Q(party=None),
        #                      name='unique_without_optional'),
        # ]
 

class workflowitems(models.Model):
    
    created_date = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)
    transitionmanager = models.OneToOneField(TransitionManager, on_delete=models.CASCADE,blank=True, null=True )
    initial_state  = models.CharField(max_length=50,default = Values.DRAFT)
    interim_state = models.CharField(max_length=50,default = Values.DRAFT)
    final_state = models.CharField(max_length=50,default = Values.DRAFT) 
    event_user = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.SET_NULL , blank=True, null=True)
    next_available_transitions =  models.JSONField(blank=True, null=True)
    from_party = models.CharField(max_length = 500  , blank = True , null = True )
    to_party = models.CharField(max_length =  500 ,  blank = True , null = True )
    action = models.CharField(max_length=25 , blank=True, null=True , default = Values.DRAFT)
    subaction = models.CharField(max_length=55 , blank=True, null=True , default = Values.DRAFT)
    previous_action = models.CharField(max_length=55 , blank=True, null=True , default = Values.DRAFT)
    record_datas = models.JSONField(blank=True, null=True)
    model_type  = models.CharField(max_length=55, blank=True, null=True)
    comments = models.CharField(max_length=500,blank=True, null=True)
    is_read = models.BooleanField(default=True,blank=True, null=True)
    final_value = models.BooleanField(default=False,blank=True, null=True)
    
    class Meta:
        verbose_name_plural = "7. WorkFlowItem"
        ordering = ['id']
    
    



# WORKEVENTS
class workevents(models.Model):

    workflowitems = models.ForeignKey(workflowitems, on_delete=models.CASCADE , related_name='WorkFlowEvents')
    action = models.CharField(max_length=25, blank=True, null=True , default = Values.DRAFT)
    subaction = models.CharField(max_length=55 , blank=True, null=True , default = Values.DRAFT)
    initial_state  = models.CharField(max_length=50 , default = Values.DRAFT)
    interim_state = models.CharField(max_length=50,default = Values.DRAFT)
    final_state = models.CharField(max_length=50,default = Values.DRAFT)
    from_party = models.CharField(max_length = 500  , blank = True , null = True )
    to_party = models.CharField(max_length = 500  , blank = True , null = True )
    event_user = models.ForeignKey(settings.AUTH_USER_MODEL,  on_delete=models.SET_NULL , blank=True, null=True)
    is_read = models.BooleanField(default=True,blank=True, null=True)
    record_datas = models.JSONField(blank=True, null=True)
    final_value = models.BooleanField(default=False,blank=True, null=True)
    comments = models.CharField(max_length=500,blank=True, null=True)
    created_date = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)
    model_type = models.CharField(max_length=55, blank=True, null=True)
    
    
    class Meta:
        verbose_name_plural = "8. WorkFlowEvent"
        ordering = ['id']



    
        








from django.conf import settings
from django.contrib.auth.models import AnonymousUser
from threading import local

USER_ATTR_NAME = getattr(settings, 'LOCAL_USER_ATTR_NAME', '_current_user')

_thread_locals = local()


def _do_set_current_user(user_fun):
    setattr(_thread_locals, USER_ATTR_NAME, user_fun.__get__(user_fun, local))


def _set_current_user(user=None):
    _do_set_current_user(lambda self: user)



class SetCurrentUser:
    def __init__(this, request):
        this.request = request

    def __enter__(this):
        _do_set_current_user(lambda self: getattr(this.request, 'user', None))

    def __exit__(this, type, value, traceback):
        _do_set_current_user(lambda self: None)


class TransitionUserMiddleware(object):

    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        with SetCurrentUser(request):
            response = self.get_response(request)
        return response


def get_current_user():
    current_user = getattr(_thread_locals, USER_ATTR_NAME, None)
    return current_user() if callable(current_user) else current_user
    

import inspect

def gets_current_users(self):
    for frame_record in inspect.stack():
        if frame_record[3]=='get_response':
            request = frame_record[0].f_locals['request']
            break
        else:
            request = None
    return request


import contextlib
from django import dispatch
from django.db.models.signals import post_save, post_delete
from .exception import DefaultsNotFoundError, PartyLengthError
from .middleware import get_current_user
from .models import States, TransitionManager, workevents, workflowitems
from django.conf import settings
from .transition import  DefaultModelValues
from .compatability import check_version_compatibility
from .middleware import get_current_user

# MAIN SIGNAL RECEIVER FOR TRANSITION HANDLING - 4/8/2022 by anand

Base_link = "https://pypi.org/project/finflo/"


# CORE LIST ITERATED FROM THE DEFAULT SETTINGS -- PRODUCTION : 27/10/2022
try:
    check_version_compatibility()
    WORK_MODEL_SENDER = list(settings.FINFLO["WORK_MODEL"])
    PARTY_MODEL_SENDER = list(settings.FINFLO["PARTY_MODEL"])
except:
    raise DefaultsNotFoundError(
        "unable to find WORK_MODEL and PARTY_MODEL in your settings.py file , \
        check the documentation for more details",
        Base_link,
    )


# CUSTOM MULTIPLE RECEIVER SIGNALS


def finflo_receiver(signal, senders, **kwargs):
    def decorator(receiver_func):
        for sender in senders:
            if isinstance(signal, (list, tuple)):
                for s in signal:
                    s.connect(receiver_func, sender=sender, **kwargs)
            else:
                signal.connect(receiver_func, sender=sender, **kwargs)
        return receiver_func
    return decorator


# MULTIPLE RECEIVER MODEL WITH MULTIPLE SENDERS

#### CREATER

## A
@finflo_receiver(post_save, senders=WORK_MODEL_SENDER)
def creates_finflo_work_models(sender, instance, created, **kwargs):
    if created:
        model = sender._meta.label_lower
        states = States.objects.get(id=1)
        obj = TransitionManager.objects.create(type=model, t_id=instance.id)
        obj2 = workflowitems.objects.create(
            transitionmanager=obj,
            model_type=model,
            record_datas=DefaultModelValues(model_type = obj.type , id = obj.t_id).get_record_datas(),
            event_user = get_current_user(),
            initial_state=states.description,
            interim_state=states.description,
            final_state=states.description,
        )
        workevents.objects.create(
            workflowitems=obj2,
            model_type=model,
            event_user = get_current_user(),
            record_datas = DefaultModelValues(model_type =  obj.type , id = obj.t_id).get_record_datas(),
            initial_state=states.description,
            interim_state=states.description,
            final_state=states.description,
        )


# POST DELETE

## A
@finflo_receiver(post_delete, senders=WORK_MODEL_SENDER)
def delete(sender, instance, **kwargs):
    TransitionManager.objects.filter(type=sender._meta.label_lower, t_id=instance.id).delete()
    


import json
from .models import (
    SignList,
    TransitionManager,
    Action,
    workevents,
    workflowitems,
    Flowmodel,
)
from .base_enum import Values
from django.core import serializers as core_serializers
from .middleware import get_current_user
from .exception import (
    ModelNotFound,
    TransitionNotAllowed,
    ReturnModelNotFound,
)
from functools import lru_cache
from django.db.models import Q
from django.conf import settings
from collections import deque
from django.apps import apps
from django.forms.models import model_to_dict
from dataclasses import dataclass 
import contextlib 
from typing import Optional



####################################################
#############       CORE     #######################
####################################################



@dataclass
class FinFlotransition:

    # BASE CONSTRUCTORS #

    t_id : int or str
    model_type: str
    source: str
    destination: str
    interim: Optional[str] = None
    action : Optional[str] = None
    party : Optional[str] = None
    from_party: Optional[str] = None
    to_party : Optional[str] = None
    record_datas : Optional[None] = None


    def __post_init__(self):
        self.action = self.action.upper() if self.action else None
        gets_return_action = self.gets_default_return_action()
        if self.action:
            if self.action == self.gets_default_return_action().description:
                self.return_transition()
            else:
                self.transition()
        else:
            self.manualtransitions()
        

    def __repr__(self):
        return f"the id is {self.t_id} and type is {self.model_type}"

    def __str__(self):
        return f"the id is {self.t_id} and type is {self.model_type}"

    # GETS THE ALL MODEL VALUE FOR TRANSITION #

    def gets_base_action(self):
        try:
            return Action.objects.get(description=self.action, model__description=self.model_type)
        except:
            raise ModelNotFound()

    # RETURN ACTION
    @staticmethod
    def gets_default_return_action():
        try:
            return Action.objects.get(id=1)
        except:
            raise ReturnModelNotFound()

    # GETS WORKFLOW MODEL ID #
    def gets_wf_item(self, model):
        return workflowitems.objects.get(transitionmanager=model)

    # GETS TRANSITION MANAGER
    def gets_base_model(self):
        try:
            return TransitionManager.objects.get(type=self.model_type, t_id=self.t_id)
        except Exception as e:
            raise ModelNotFound() from e

    def sign_reset(self, overall_model):
        overall_model[0].sub_sign = 0
        overall_model[0].in_progress = False
        overall_model[0].save()

    

    def get_record_datas(self):
        try:
            overall_model = self.gets_base_model()
            gets_model = apps.get_model(overall_model.type)
            query_data = gets_model.objects.filter(id = overall_model.t_id)
            base_serialized_Data = core_serializers.serialize("json", query_data)
            if query_data.exists():
                return {"values": json.loads(base_serialized_Data)}
        except Exception:
            return {"values": None}



    # GETS ALL ACTION
    def gets_all_models(self):
        try:
            gets_model = TransitionManager.objects.get(type__icontains=self.model_type, t_id=self.t_id)
            gets_flows = Flowmodel.objects.get(description__icontains=self.model_type)
            gets_action = Action.objects.get(
                Q(model=gets_flows.id) | Q(model=None),
                description=self.action,
                party__id = self.party or None,
            )
            gets_wf = self.gets_wf_item(gets_model.id)
            sign_lists = []
            sub_action_list = []
            try:
                for item in SignList.objects.all():
                    sign_lists.append(item.name)
                    sub_action_list.append(item.sub_action_name)
                    if item.name == gets_action.stage_required.name:
                        break
                next_avail_trans = sign_lists[gets_model.sub_sign :]
                next_avail_trans_value = deque(next_avail_trans)
                next_avail_trans_value.popleft()
                next_avail_Transition = {"values": list(next_avail_trans_value)}
            except Exception:
                next_avail_Transition = None
            return (
                gets_model,
                gets_action,
                gets_flows,
                gets_wf,
                sign_lists,
                next_avail_Transition,
                sub_action_list
            )
        except:
            raise ModelNotFound()


    def return_transition(self):

        overall_model = self.gets_all_models()
        obj, created = workflowitems.objects.update_or_create(
            transitionmanager=overall_model[0] or overall_model[0].id,
            defaults={
                "initial_state": overall_model[1].from_state.description
                if overall_model[1].from_state
                else overall_model[3].initial_state,
                "interim_state": overall_model[1].to_state.description,
                "final_state": overall_model[1].to_state.description,
                "action": self.action,
                "record_datas": self.get_record_datas(),
                "subaction": self.action,
                "previous_action": overall_model[3].action,
                "next_available_transitions": None,
                "model_type": self.model_type,
                "event_user": get_current_user(),
                "final_value": True,
                "from_party": overall_model[1].from_party or self.from_party,
                "to_party": overall_model[1].to_party or self.to_party,
            },
        )
        workevents.objects.create(
            workflowitems=overall_model[3],
            event_user=get_current_user(),
            initial_state=overall_model[1].from_state.description
            if overall_model[1].from_state
            else overall_model[3].initial_state,
            final_value=True,
            record_datas=self.get_record_datas(),
            interim_state=overall_model[1].to_state.description,
            final_state=overall_model[1].to_state.description,
            action=self.action,
            subaction=self.action,
            model_type=self.model_type,
            from_party=overall_model[1].from_party or self.from_party,
            to_party=overall_model[1].to_party or self.to_party,
        )
        self.sign_reset(overall_model)

    




    # MANUAL TRANSITION WITH SOURCE , INTERIM , AND destination STATES

    def gets_default_model(self):
        obj , created = TransitionManager.objects.get_or_create(t_id =  self.t_id , type = self.model_type)
        return obj


    def manualtransitions(self):
        try:
            workflowitems.objects.filter(
                transitionmanager = self.gets_default_model()).update(
                initial_state =  self.source,
                interim_state =  self.interim,
                final_state =  self.destination,
                next_available_transitions = None,
                record_datas = self.get_record_datas(),
                model_type = self.model_type,
                event_user =  get_current_user(),
                from_party = self.from_party,
                to_party = self.to_party,
                final_value = True,
            )
            workevents.objects.create(
            workflowitems = workflowitems.objects.get(transitionmanager = self.gets_default_model()),
            event_user=get_current_user(),
            initial_state=self.source,
            interim_state=self.interim,
            final_state=self.destination,
            model_type=self.model_type,
            record_datas =  self.get_record_datas(),
            from_party=self.from_party,
            to_party=self.to_party,
            final_value=True,
            )
        except Exception:
            data = workflowitems.objects.create(
                transitionmanager = self.gets_default_model(),
                initial_state =  self.source,
                interim_state =  self.interim,
                final_state =  self.destination,
                next_available_transitions = None,
                record_datas = self.get_record_datas(),
                model_type = self.model_type,
                event_user =  get_current_user(),
                from_party = self.from_party,
                to_party = self.to_party,
                final_value = True,
            )
            workevents.objects.create(
            workflowitems = data ,
            event_user=get_current_user(),
            initial_state=self.source,
            interim_state=self.interim,
            final_state=self.destination,
            model_type=self.model_type,
            record_datas= self.get_record_datas(),
            from_party=self.from_party,
            to_party=self.to_party,
            final_value=True,
            )
            # data = workflowitems.objects.get(transitionmanager = self.gets_default_model())

            
        # except Exception as e:
        #     raise TransitionNotAllowed() from e


    ## CORE TRANSITION ###
    def gets_subaction_name(self):
        with contextlib.suppress(Exception):
            overall_model = self.gets_all_models()
            data =  overall_model[6][overall_model[0].sub_sign - 1]
            return "" if data == Values.INITIAL_SIGN  else data
        return None

    def transition(self):

        overall_model = self.gets_all_models()

        if overall_model[0] is None:
            raise TransitionNotAllowed()
        try:
            ws = workflowitems.objects.update_or_create(
                transitionmanager=overall_model[0] or overall_model[0].id,
                defaults={
                    "initial_state": overall_model[1].from_state.description,
                    "interim_state": overall_model[4][
                        1 + overall_model[0].sub_sign
                    ], 
                    "final_state": overall_model[1].to_state.description,
                    "next_available_transitions": overall_model[5],
                    "action": self.action,
                    "subaction": overall_model[6][overall_model[0].sub_sign],
                    "model_type": self.model_type,
                    "record_datas": self.get_record_datas(),
                    "event_user": get_current_user(),
                    "from_party": overall_model[1].from_party or self.from_party,
                    "to_party": overall_model[1].to_party or self.to_party,
                },
            )
            workevents.objects.create(
                workflowitems=overall_model[3],
                event_user=get_current_user(),
                initial_state=overall_model[1].from_state.description,
                interim_state=overall_model[4][1 + overall_model[0].sub_sign],
                record_datas=self.get_record_datas(),
                final_state=overall_model[1].to_state.description,
                action=self.action,
                subaction=overall_model[6][overall_model[0].sub_sign],
                model_type=self.model_type,
                from_party=overall_model[1].from_party or self.from_party,
                to_party=overall_model[1].to_party or self.to_party,
            )
            overall_model[0].sub_sign += 1
            overall_model[0].in_progress = True
            overall_model[0].save()
        except Exception:
            ws = workflowitems.objects.update_or_create(
                transitionmanager=overall_model[0] or overall_model[0].id,
                defaults={
                    "initial_state": overall_model[1].from_state.description,
                    "interim_state": overall_model[1].to_state.description,
                    "final_state": overall_model[1].to_state.description,
                    "action": self.action,
                    "next_available_transitions": None,
                    "subaction": self.action,
                    "model_type": self.model_type,
                    "record_datas": self.get_record_datas(),
                    "event_user": get_current_user(),
                    "from_party": overall_model[1].from_party or self.from_party,
                    "to_party": overall_model[1].to_party or self.to_party,
                    "final_value": True,
                },
            )
            workevents.objects.create(
                workflowitems=overall_model[3],
                event_user=get_current_user(),
                initial_state=overall_model[1].from_state.description,
                interim_state=overall_model[1].to_state.description,
                record_datas=self.get_record_datas(),
                final_state=overall_model[1].to_state.description,
                action=self.action,
                subaction=self.action,
                model_type=self.model_type,
                from_party=overall_model[1].from_party or self.from_party,
                to_party=overall_model[1].to_party or self.to_party,
                final_value=True,
            )
            # self.sign_reset(overall_model)
            overall_model[0].sign_reset

    
    



@dataclass
class DefaultModelValues:

    model_type : str
    id : str
    
    def get_record_datas(self):
        try:
            gets_model = apps.get_model(self.model_type)
            query_data = gets_model.objects.filter(id=self.id)
            base_serialized_Data = core_serializers.serialize("json", query_data)
            if query_data.exists():
                return {"values": json.loads(base_serialized_Data)}
        except Exception:
            return {"values": None}


# Generated by Django 3.2.5 on 2022-10-14 10:50

import contextlib
from django.conf import settings
from django.db import migrations



# RUN ON MIGRATIONS  -- remigrate if work_model is updated 


def print_migration_message(message):
    BOLD = "\033[1m"
    GREEN = "\033[92m"
    RESET = "\033[0m"
    return f"{BOLD}{GREEN}{message}{RESET}"



def update_flow_model(apps, schema_editor):
    flowmodel = apps.get_model("finflo.Flowmodel")
    try:
        a = settings.FINFLO['WORK_MODEL']
        db_alias = schema_editor.connection.alias
        for model in a:
            print(print_migration_message("Applying Finflo migrations..."))
            flowmodel.objects.using(db_alias).update_or_create(description = model.lower())
    except:
        raise LookupError("Couldn't find work_model in settings.py check documentation")





    
class Migration(migrations.Migration):

    dependencies = [
        ('finflo', '0001_initial'),
    ]

    operations = [
        migrations.RunPython(update_flow_model),
    ]


from enum import Enum 



""" default enum values cannot be deleted from admin """ 
""" hint : can be modified """
class Values(Enum):
    DRAFT = "DRAFT"
    RETURN = "RETURN"
    INITIAL_SIGN = "initial sign"
    AWAITING_SIGN_A = "Awaiting Sign A"
    AWAITING_SIGN_B = "Awaiting Sign B"
    AWAITING_SIGN_C = "Awaiting Sign C"
    SIGN_A  = "sign a"
    SIGN_B = "sign b"
    SIGN_C = "sign c"

    
    def __repr__(self):
        return f"{self.value}"
    
    def __str__(self):
        return f"{self.value}"

    def __eq__(self, other):
        return self.value == other if isinstance(other, str) else super().__eq__(other)

    




# IMP_NOTE:  THE DRAFT IS DEFAULT VALUE FOR A TRANSITION STARTED PACK 